/*
 * Partially implemented disassembler for Samsung SSP160x processors
 * (C) Copyright 2007-2008, Grazvydas "notaz" Ignotas
 */

#include "ssp.hpp"

#define LSB(x) ((x)&0xff)
#define LSHB(x) ((x)&0xf)
#define MSHB(x) (((x)>>4)&0xf)

//----------------------------------------------------------------------
// analyze an instruction
int idaapi ana(void)
{
  int bit_f;
  ushort op;
  
  op = ua_next_word();
  bit_f = (op >> 8) & 1;

  // analyze it!
  switch (op >> 9)
  {
    // OP a, s
    case 0x10: // sub
      cmd.itype = SSP_sub;
      goto op_a_s;
    case 0x30: // cmp
      cmd.itype = SSP_cmp;
      goto op_a_s;
    case 0x40: // add
      cmd.itype = SSP_add;
      goto op_a_s;
    case 0x50: // and
      cmd.itype = SSP_and;
      goto op_a_s;
    case 0x60: // or
      cmd.itype = SSP_or;
      goto op_a_s;
    case 0x70: // eor
      cmd.itype = SSP_eor;
    op_a_s:
      if (op & 0x1f0) goto unknown;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = rA;
      cmd.Op2.type = o_reg;
      cmd.Op2.reg = LSHB(op);
      break;

    // OP a, (ri)
    case 0x11: // sub
      cmd.itype = SSP_sub;
      goto op_a_i_ri;
    case 0x31: // cmp
      cmd.itype = SSP_cmp;
      goto op_a_i_ri;
    case 0x41: // add
      cmd.itype = SSP_add;
      goto op_a_i_ri;
    case 0x51: // and
      cmd.itype = SSP_and;
      goto op_a_i_ri;
    case 0x61: // or
      cmd.itype = SSP_or;
      goto op_a_i_ri;
    case 0x71: // eor
      cmd.itype = SSP_eor;
    op_a_i_ri:
      if (op & 0xf0) goto unknown;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = rA;
      cmd.Op2.type = o_idpspec1; // ri|mod
      cmd.Op2.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op2.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op2.specflag2 = 1; // indirection counter
      break;

    // OP a, adr
    case 0x13: // sub
      cmd.itype = SSP_sub;
      goto op_a_adr;
    case 0x33: // cmp
      cmd.itype = SSP_cmp;
      goto op_a_adr;
    case 0x43: // add
      cmd.itype = SSP_add;
      goto op_a_adr;
    case 0x53: // and
      cmd.itype = SSP_and;
      goto op_a_adr;
    case 0x63: // or
      cmd.itype = SSP_or;
      goto op_a_adr;
    case 0x73: // eor
      cmd.itype = SSP_eor;
    op_a_adr:
      // if (bit_f) goto unknown; // ??
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = rA;
      cmd.Op2.type = o_imm;
      cmd.Op2.dtyp = dt_word;
      cmd.Op2.value = op & 0x1ff;
      break;

    // OP a, imm
    case 0x14: // subi a, imm
      cmd.itype = SSP_subi;
      goto op_a_imm;
    case 0x34: // cmpi a, imm
      cmd.itype = SSP_cmpi;
      goto op_a_imm;
    case 0x44: // addi a, imm
      cmd.itype = SSP_addi;
      goto op_a_imm;
    case 0x54: // andi a, imm
      cmd.itype = SSP_andi;
      goto op_a_imm;
    case 0x64: // ori  a, imm
      cmd.itype = SSP_ori;
      goto op_a_imm;
    case 0x74: // eori a, imm
      cmd.itype = SSP_eori;
    op_a_imm:
      if (op & 0x1ff) goto unknown;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = rA;
      cmd.Op2.type = o_imm;
      cmd.Op2.dtyp = dt_word;
      cmd.Op2.value = ua_next_word();
      break;

    // OP a, ((ri))
    case 0x15: // sub
      cmd.itype = SSP_sub;
      goto op_a_ii_ri;
    case 0x35: // cmp
      cmd.itype = SSP_cmp;
      goto op_a_ii_ri;
    case 0x45: // add
      cmd.itype = SSP_add;
      goto op_a_ii_ri;
    case 0x55: // and
      cmd.itype = SSP_and;
      goto op_a_ii_ri;
    case 0x65: // or
      cmd.itype = SSP_or;
      goto op_a_ii_ri;
    case 0x75: // eor
      cmd.itype = SSP_eor;
    op_a_ii_ri:
      if (op & 0xf0) goto unknown;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = rA;
      cmd.Op2.type = o_idpspec1; // ri|mod
      cmd.Op2.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op2.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op2.specflag2 = 2; // indirection counter
      break;

    // OP a, ri
    case 0x19: // sub
      cmd.itype = SSP_sub;
      goto op_a_ri;
    case 0x39: // cmp
      cmd.itype = SSP_cmp;
      goto op_a_ri;
    case 0x49: // add
      cmd.itype = SSP_add;
      goto op_a_ri;
    case 0x59: // and
      cmd.itype = SSP_and;
      goto op_a_ri;
    case 0x69: // or
      cmd.itype = SSP_or;
      goto op_a_ri;
    case 0x79: // eor
      cmd.itype = SSP_eor;
    op_a_ri:
      if (op & 0xf0) goto unknown;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = rA;
      cmd.Op2.type = o_idpspec1; // ri|mod
      cmd.Op2.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op2.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op2.specflag2 = 0; // indirection counter
      break;

    // OP (rj), (ri), b
    case 0x0b:
    case 0x1b: // ???
      cmd.itype = SSP_mpys;
      goto op_rj_ri_b;
    case 0x4b:
      cmd.itype = SSP_mpya;
      goto op_rj_ri_b;
    case 0x5b:
      cmd.itype = SSP_mld;
    op_rj_ri_b:
      cmd.Op1.type = o_idpspec1; // rij|mod
      cmd.Op1.specval = ((op >> 4) & 3) | 4; // rj index
      cmd.Op1.specflag1 = (op >> 6) & 3; // modifier
      cmd.Op1.specflag2 = 1;     // ind. level
      cmd.Op2.type = o_idpspec1; // rij|mod
      cmd.Op2.specval = op & 3;  // ri index
      cmd.Op2.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op2.specflag2 = 1;     // ind. level
      cmd.Op3.type = o_idpspec4; // 'b' operand
      cmd.Op3.specval = bit_f;
      break;

    // OP simm
    case 0x1c: // subi simm
      cmd.itype = SSP_subi_simm;
      goto op_simm;
    case 0x3c: // cmpi simm
      cmd.itype = SSP_cmpi_simm;
      goto op_simm;
    case 0x4c: // addi simm
      cmd.itype = SSP_addi_simm;
      goto op_simm;
    case 0x5c: // andi simm
      cmd.itype = SSP_andi_simm;
      goto op_simm;
    case 0x6c: // ori  simm
      cmd.itype = SSP_ori_simm;
      goto op_simm;
    case 0x7c: // eori simm
      cmd.itype = SSP_eori_simm;
    op_simm:
      if (bit_f) goto unknown;
      cmd.Op1.type = o_imm;
      cmd.Op1.dtyp = dt_byte;
      cmd.Op1.value = op & 0xff;
      break;

    case 0x00:
      if (bit_f) goto unknown;
      if (LSB(op) == 0x00)
        cmd.itype = SSP_nop;
      else if (LSB(op) == 0x65)
        cmd.itype = SSP_ret;
      else {
        cmd.itype = SSP_ld;
        cmd.Op1.type = o_reg;
        cmd.Op1.reg = MSHB(op);
        cmd.Op2.type = o_reg;
        cmd.Op2.reg = LSHB(op);
      }
      break;

    case 0x01: // ld d, (ri) // MAME doubts
      cmd.itype = SSP_ld;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = MSHB(op);
      cmd.Op2.type = o_idpspec1;  // ri|mod
      cmd.Op2.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op2.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op2.specflag2 = 1; // indirection counter
      break;

    case 0x02: // ld (ri), s
      cmd.itype = SSP_ld;
      cmd.Op1.type = o_idpspec1;  // ri|mod
      cmd.Op1.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op1.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op1.specflag2 = 1; // indirection counter
      cmd.Op2.type = o_reg;
      cmd.Op2.reg = MSHB(op);
      break;

    case 0x03: // ld a, adr
      cmd.itype = SSP_ld;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = rA;
      cmd.Op2.type = o_imm;
      cmd.Op2.dtyp = dt_word;
      cmd.Op2.value = op & 0x1ff;
      break;

    case 0x04: // ldi d, imm
      if (op & 0x10f) goto unknown;
      cmd.itype = SSP_ldi;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = MSHB(op);
      cmd.Op2.type = o_imm;
      cmd.Op2.dtyp = dt_word;
      cmd.Op2.value = ua_next_word();
      if (cmd.Op1.reg == rPC) {
        cmd.Op2.type = o_near;
        cmd.Op2.addr = cmd.Op2.value << 1;
	cmd.Op2.value = 0;
      }
      break;

    case 0x05: // ld d, ((ri))
      cmd.itype = SSP_ld;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = MSHB(op);
      cmd.Op2.type = o_idpspec1; // ri|mod
      cmd.Op2.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op2.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op2.specflag2 = 2; // indirection counter
      break;

    case 0x06: // ldi (ri), imm
      cmd.itype = SSP_ldi;
      cmd.Op1.type = o_idpspec1;  // ri|mod
      cmd.Op1.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op1.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op1.specflag2 = 1; // indirection counter
      cmd.Op2.type = o_imm;
      cmd.Op2.dtyp = dt_word;
      cmd.Op2.value = ua_next_word();
      break;

    case 0x07: // ld adr, a // MAME doubts
      cmd.itype = SSP_ld;
      cmd.Op1.type = o_imm;
      cmd.Op1.dtyp = dt_word;
      cmd.Op1.value = op & 0x1ff;
      cmd.Op2.type = o_reg;
      cmd.Op2.reg = rA;
      break;

    case 0x09: // ld d, ri
      cmd.itype = SSP_ld;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = MSHB(op);
      cmd.Op2.type = o_idpspec1;  // ri|mod
      cmd.Op2.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op2.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op2.specflag2 = 0; // indirection counter
      break;

    case 0x0a: // ld ri, s
      cmd.itype = SSP_ld;
      cmd.Op1.type = o_idpspec1;  // ri|mod
      cmd.Op1.specval = (bit_f << 2) | (op & 3); // rij index
      cmd.Op1.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op1.specflag2 = 0; // indirection counter
      cmd.Op2.type = o_reg;
      cmd.Op2.reg = MSHB(op);
      break;

    // ldi ri, simm
    case 0x0c:
    case 0x0d:
    case 0x0e:
    case 0x0f:
      cmd.itype = SSP_ldi;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = 0x20 | ((op >> 8) & 7); // ri
      cmd.Op2.type = o_imm;
      cmd.Op2.dtyp = dt_byte;
      cmd.Op2.value = op & 0xff;
      break;

    // mset reg, (ri), s // ???
/*
    case 0x1a:
    case 0x1b:
      cmd.itype = SSP_mset;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = (op >> 5) & 0x1f;
      cmd.Op2.type = o_idpspec1;  // ri|mod
      cmd.Op2.specval = op & 3;   // ri index
      cmd.Op2.specflag1 = (op >> 2) & 3; // modifier
      cmd.Op2.specflag2 = 1; // indirection counter
      cmd.Op3.type = o_idpspec4; // 's' operand
      cmd.Op3.specval = (op >> 4) & 1;
      break;
*/
    case 0x24: // call cond, addr
    case 0x26: // bra  cond, addr
      cmd.itype = (op >> 9) == 0x24 ? SSP_call : SSP_bra;
      cmd.Op1.type = o_idpspec0; // guessed flag check
      cmd.Op1.specval = MSHB(op);
      cmd.Op1.specflag1 = bit_f;
      cmd.Op2.type = o_near;
      cmd.Op2.dtyp = dt_word;
      cmd.Op2.addr = ua_next_word() << 1;
      break;

    case 0x25: // ld d, (a) // based on Tasco's docs, not MAME code
      if (op & 0x10f) goto unknown;
      cmd.itype = SSP_ld;
      cmd.Op1.type = o_reg;
      cmd.Op1.reg = MSHB(op);
      cmd.Op2.type = o_idpspec2;  // (a)
      cmd.Op2.dtyp = dt_word;
      break;

    case 0x48: // mod cond, op
      if (op & 0x008) goto unknown;
      cmd.itype = SSP_mod;
      cmd.Op1.type = o_idpspec0;  // flag check
      cmd.Op1.specval = MSHB(op);
      cmd.Op1.specflag1 = bit_f;
      cmd.Op2.type = o_idpspec3;  // mod operand
      cmd.Op2.specval = op & 7;
      break;

    case 0x4a: // mod f, op
      if (op & 0x1f0) goto unknown;
      cmd.itype = SSP_mod;
      cmd.Op1.type = o_idpspec5;  // "mod f" stuff
      cmd.Op1.specval = 16;       // "f"
      cmd.Op2.type = o_idpspec5;
      cmd.Op2.specval = op & 0xf;
      break;


    unknown:
    default:
      cmd.itype = SSP_unknown;
      cmd.auxpref = op;
      break;
  }

  return cmd.size;
}
